/*
 Copyright (C)2004 Dimax ( http://www.xdimax.com )

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
*/
  
/*
*-------------------------------------------------------------------
* Description: i2cbrdg main device routines module
*-------------------------------------------------------------------								 
* Version: $Id: dev_rout.c,v 1.1 2005/03/08 15:57:47 dmitry_kulik Exp $									   
*-------------------------------------------------------------------
*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include <usb.h>
//#include "i2cbridge.h"
#include "dev_rout.h"


//-----------------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------------

#define DIMAX_VENDOR    0x0abf
#define DIMAX_PRODUCT   0x3370
#define DIMAX_IF        0x00
#define DIMAX_IN_EP     0x84
#define DIMAX_OUT_EP    0x02

#define DIMAX_TIMEOUT   100     // ms

int debug=0;

//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------
static int usb_init_done;

//-----------------------------------------------------------------------------
// Local Function Definition
//-----------------------------------------------------------------------------
static U2C_RESULT u2c_libusb_rc( int rc );


BYTE dr_get_device_count( void )
{
    BYTE    dev_count=0;
    struct  usb_bus *bus;
    struct  usb_device *dev;

    if( ! usb_init_done )
    {
        usb_init();
        usb_init_done=1;
    }

    usb_find_busses();
    usb_find_devices();
    bus=usb_get_busses();

    for( ; bus; bus=bus->next )
        for( dev=bus->devices; dev; dev=dev->next )
        {
            if ( dev->descriptor.idVendor  == DIMAX_VENDOR &&
                 dev->descriptor.idProduct == DIMAX_PRODUCT ) 
            {
                dev_count++;
            }
        }
        
    return dev_count;
}

HANDLE dr_open_device(BYTE nDevice)
{
	BYTE    devc=0;
	int     found=0;
	struct  usb_bus *bus;
	struct  usb_device *dev=0;
	HANDLE  hDevice;

	if( ! usb_init_done )
	{
		usb_init();
		usb_init_done=1;
	}

	usb_find_busses();
	usb_find_devices();
	bus=usb_get_busses();

	while(bus && !found)
	{
		for( dev=bus->devices; dev; dev=dev->next )
		{
			if ( dev->descriptor.idVendor  == DIMAX_VENDOR &&
				 dev->descriptor.idProduct == DIMAX_PRODUCT ) 
			{
				if( devc == nDevice )
				{
					found = 1;
					break;
				}
				devc++;
			}
		}
		bus = bus->next;
	}

	if( ! dev )
		return INVALID_HANDLE_VALUE;

	/* 
	* Open device file 
	* You should have sufficient permisions to do that
	*/
	hDevice = usb_open(dev);
	if( ! hDevice )
		return INVALID_HANDLE_VALUE;

	return hDevice;
}

U2C_RESULT dr_close_device(HANDLE hDevice)
{
	usb_close( hDevice );
	return U2C_SUCCESS;
}

U2C_RESULT dr_is_handle_valid(HANDLE hDevice)
{
	if( hDevice )
		return U2C_SUCCESS;
	else
		return U2C_HARDWARE_NOT_FOUND;
}

U2C_RESULT dr_get_driver_version(HANDLE hDevice, PU2C_VERSION_INFO pVersion)
{
    return U2C_NOT_IMPLEMENTED;
}

//-----------------------------------------------------------------------------
// Local Functions 
//-----------------------------------------------------------------------------

/*
* Convert errno codes returned by libusb to U2C_RESULT
*/
static U2C_RESULT u2c_libusb_rc( int rc )
{
    if( rc )
        switch( rc )
        {            
        case EBUSY:
            return U2C_DEVICE_BUSY;
        case ENOMEM:
            return U2C_MEMORY_ERROR;
        default:
            return U2C_UNKNOWN_ERROR;
        }
    else
        return U2C_SUCCESS;
}   

/*
* Convert FW status codes to U2C_RESULT
*/
U2C_RESULT u2c_fw_rc( int rc )
{
    if( rc )
        switch( rc )
        {            
        case 1:
            return U2C_TRANSACTION_FAILED;
        case 2:
            return U2C_SLAVE_OPENNING_FOR_WRITE_FAILED;
        case 3:
            return U2C_SLAVE_OPENNING_FOR_READ_FAILED;
        case 4:
            return U2C_SENDING_MEMORY_ADDRESS_FAILED;
        case 5:
            return U2C_SENDING_DATA_FAILED;
        case 6:
            return U2C_NOT_IMPLEMENTED;
        case 7:
            return U2C_NO_ACK;            
        default:
            return U2C_UNKNOWN_ERROR;
        }
    else
        return U2C_SUCCESS;
}   

/*
* Dump buffer
*/
static void dump_buf( char* title, char* buf, int size )
{
    int     i;

    if( title )
        printf( title );
    for( i=0; i<size; i++ )
        printf( "%02x ", ((unsigned int)buf[i])&0xFF );    
    printf("\n");
}    

//-----------------------------------------------------------------------------
// U2C Transaction Core Function 
//-----------------------------------------------------------------------------
U2C_RESULT dr_u2c_transaction( HANDLE hDevice, PDEVICE_REQUEST req )
{
    int     rc=0;
    sigset_t set, oldset;
    int     i;

/* 
 After interface claimed no other process can access 
 same interface, so this is good for locking     
*/
    
    /* Claim Interface */
    rc = usb_claim_interface( hDevice, DIMAX_IF );
    if( rc )
        return u2c_libusb_rc(rc);
    
/*
 To be sure signal will not kill process between out and in bulk 
 transactions postpone signals. (This will not stop SIGKILL and SIGSTOP)
*/     
    
    /* Block signals to make transaction atomic */
    memset( &set, 0xFF, sizeof(sigset_t) );
    sigprocmask(SIG_BLOCK, &set, &oldset);

    /* Bulk out */
    if( debug )
        dump_buf( "Bulk Out: ", req->OutputBuffer, req->OutputBufferLength ); 
    rc = usb_bulk_write( hDevice, DIMAX_OUT_EP, 
                req->OutputBuffer, req->OutputBufferLength, DIMAX_TIMEOUT );
    if( rc != req->OutputBufferLength )
    {        
        if( debug )
            printf( "usb_bulk_write failed rc=%d\n", rc );
        rc = U2C_TRANSACTION_FAILED;
        goto cleanup;
    }        

    /* Bulk in */
    rc = usb_bulk_read( hDevice, DIMAX_IN_EP, 
                req->InputBuffer, MAX_INPUT_BUFFER_LENGTH, DIMAX_TIMEOUT );
 
    if (req->InputBufferLength == 0)
       req->InputBufferLength = rc;
    if(( rc != req->InputBufferLength ) || (rc == 0))
    {        
        if( debug )
            printf( "usb_bulk_read failed rc=%d\n", rc );
        rc = U2C_TRANSACTION_FAILED;
        goto cleanup;
    }        
    if( debug )
        dump_buf( "Bulk In: ", req->InputBuffer, req->InputBufferLength ); 

    /* Check Status. Status is returned by fw in last byte */
    if( req->InputBuffer[req->InputBufferLength-1] )
        rc = u2c_fw_rc( req->InputBuffer[req->InputBufferLength-1] );
    else
        rc = U2C_SUCCESS;

    
cleanup:    
    /* Unblock signals which was not blocked */
    for( i=0; i<sizeof(sigset_t); i++ )
        ((char*)(&set))[i]=~((char*)(&oldset))[i];
    sigprocmask(SIG_UNBLOCK, &set, &oldset);

    /* Release Interface */
    usb_release_interface( hDevice, DIMAX_IF ); 
    return rc;
}


U2C_RESULT dr_u2c_transaction_pack( HANDLE hDevice, PDEVICE_REQUEST req, int reqc )
{
    int     rc=0;
    sigset_t set, oldset;
    int     i,offs;
    char    out_buf[128];
    char    in_buf[MAX_INPUT_BUFFER_LENGTH];

/* 
 After interface claimed no other process can access 
 same interface, so this is good for locking     
*/
    
    /* Claim Interface */
    rc = usb_claim_interface( hDevice, DIMAX_IF );
    if( rc )
        return u2c_libusb_rc(rc);
    
/*
 To be sure signal will not kill process between out and in bulk 
 transactions postpone signals. (This will not stop SIGKILL and SIGSTOP)
*/     
    
    /* Block signals to make transaction atomic */
    memset( &set, 0xFF, sizeof(sigset_t) );
    sigprocmask(SIG_BLOCK, &set, &oldset);

    /* 
     * Bulk out 
    */

    /* Assemble the request */
    for( i=0,offs=0; i<reqc; i++ )
    {
        /* 
         * Request is 128 bytes limited to escape deadlock
        */
        if( offs + req[i].OutputBufferLength > 128 )
            return U2C_BAD_PARAMETER; 
        memcpy( out_buf+offs, req[i].OutputBuffer, req[i].OutputBufferLength );
        offs += req[i].OutputBufferLength;
    }        
            
    /* Send request */
    rc = usb_bulk_write( hDevice, DIMAX_OUT_EP, out_buf, offs, DIMAX_TIMEOUT );
    if( rc != offs )
    {        
        if( debug )
            printf( "usb_bulk_write pack failed rc=%d\n", rc );
        rc = U2C_TRANSACTION_FAILED;
        goto cleanup;
    }        
    
    /* 
     * Bulk in 
    */
    for( i=0; i<reqc; i++ )
    {
        rc = usb_bulk_read( hDevice, DIMAX_IN_EP, in_buf, MAX_INPUT_BUFFER_LENGTH, DIMAX_TIMEOUT );
        /*
         * Bulk result will be returned in InputBufferLength
         * Positive value will represent actual data size
         * Negative value is error code
         * Upper level function should check FW status in the last byte
        */
        req[i].InputBufferLength = rc;
        if( rc > 0 )
            memcpy( req[i].InputBuffer, in_buf, rc );
    }
    rc=0;
    
cleanup:    
    /* Unblock signals which was not blocked */
    for( i=0; i<sizeof(sigset_t); i++ )
        ((char*)(&set))[i]=~((char*)(&oldset))[i];
    sigprocmask(SIG_UNBLOCK, &set, &oldset);

    /* Release Interface */
    usb_release_interface( hDevice, DIMAX_IF ); 
    return rc;
}


